﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using Draw = System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Reflection;
using System.IO;
using belintesa.FastCode;
using belintesa.Imagine;
using Imagine = belintesa.Imagine;
using belintesa.Imagine.Filtering;

namespace KidPaint
{
    public partial class FM : Form
    {
        #region Types
        private class UndoInfo
        {
            public Imagine.Color Color { get; set; }
            public Imagine.Point Point { get; set; }

            public UndoInfo( Imagine.Color color, int x, int y )
                : this( color, new Imagine.Point( x, y ) ) { }

            public UndoInfo( Imagine.Color color, Imagine.Point point )
            {
                Color = color;
                Point = point;
            }
        }
        #endregion

        #region Fields
        private static readonly Random Rand = new Random( (int)(DateTime.Now.Ticks % int.MaxValue) );
        private static readonly string ImageFolder = Path.Combine( Application.StartupPath, "img" );
        private static List<string> ImageFiles = Directory.GetFiles( ImageFolder ).ToList();
        private static List<SolidBrush> Brushes = new List<SolidBrush>();

        private string ImageFile;
        private Picture Picture;
        private Imagine.Color PaintColor = Imagine.Color.Transparent;
        private Stack<UndoInfo> History = new Stack<UndoInfo>();
        #endregion

        #region Methods
        static FM()
        {
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 76, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 76, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 76, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 76, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 76, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 76, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 32, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 32, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 32, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 32, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 32, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 32, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 32, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 32, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 32, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 32, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 32, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 32, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 32, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 32, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 32, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 32, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 32, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 32, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 32, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 32, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 32, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 32, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 32, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 32, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 32, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 32, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 32, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 32, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 32, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 32, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 32, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 32, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 32, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 32, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 32, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 32, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 76, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 76, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 76, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 76, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 76, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 76, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 76, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 76, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 76, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 76, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 76, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 76, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 76, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 76, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 76, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 76, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 76, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 76, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 76, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 76, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 76, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 76, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 76, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 76, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 76, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 76, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 76, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 76, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 76, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 76, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 120, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 120, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 120, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 120, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 120, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 120, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 164, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 164, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 164, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 164, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 164, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 164, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 208, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 208, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 208, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 208, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 208, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 208, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 252, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 252, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 252, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 252, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 252, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 32, 252, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 252, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 252, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 252, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 252, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 252, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 76, 252, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 252, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 252, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 252, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 252, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 252, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 120, 252, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 252, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 252, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 252, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 252, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 252, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 164, 252, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 252, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 252, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 252, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 252, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 252, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 208, 252, 252 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 252, 32 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 252, 76 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 252, 120 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 252, 164 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 252, 208 ) ) );
            Brushes.Add( new SolidBrush( Draw.Color.FromArgb( 252, 252, 252 ) ) );

            //Brushes.Sort( (b1, b2) => b1.GetHash().CompareTo( b2.GetHash() ) );
        }

        public FM()
        {
            InitializeComponent();
            ChangePicture();

            int cnt = 0;

            Brushes.ForEach(
                brush =>
                {
                    Button sample = new Button();

                    sample.FlatStyle = FlatStyle.Popup;
                    sample.BackColor = brush.Color;
                    sample.Size = new Size( 40, 40 );
                    sample.Padding = sample.Margin = new Padding( 0 );
                    sample.Click += delegate( object s, EventArgs e ) { PaintColor = sample.BackColor; };
                    sample.TabIndex = cnt++;

                    BrushPanel.Controls.Add( sample );
                } );

            OnResize( null );
        }

        private void ChangePicture()
        {
            ImageFile = ImageFiles[Rand.Next( ImageFiles.Count )];
            SetPicture( Image.FromFile( ImageFile ) );
        }

        private void SetPicture( Image image )
        {
            History.Clear();
            Picture = image;
            image.Dispose();

            for ( int x = 0; x < Picture.Width; x++ )
            {
                for ( int y = 0; y < Picture.Height; y++ )
                {
                    var color = Picture.ColorPoints[x, y];

                    bool clean =
                        color.Brightness > 0.15
                        && Math.Abs( color.R - color.B ) < 4
                        && Math.Abs( color.R - color.G ) < 4
                        && Math.Abs( color.G - color.B ) < 4;
                    
                    if ( clean )
                        Picture.ColorPoints[x, y] = Imagine.Color.White;
                }
            }

            ImageBox.Image = Picture;
        }

        private void Fill( int x, int y, Imagine.Color color )
        {
            int white = Picture.ColorPoints[x, y].ARGB;

            if ( white == Imagine.Color.Black.ARGB || white == color.ARGB )
                return;

            Fill( x, y, color, new Area( 0, 0, Picture.Width, Picture.Height ) );
            ImageBox.Image.Dispose();
            ImageBox.Image = Picture;
        }

        private void Fill( int x, int y, Imagine.Color color, Area area )
        {
            int white = Picture.ColorPoints[x, y].ARGB;

            int sx = area.X;
            int sy = area.Y;
            int fx = sx + area.Width;
            int fy = sy + area.Height;

            Queue<Imagine.Point> points = new Queue<Imagine.Point>( Picture.Width * Picture.Height / 16 );
            points.Enqueue( new Imagine.Point( x, y ) );
            Picture.ColorPoints[x, y] = color;

            while ( points.Count > 0 )
            {
                Imagine.Point p = points.Dequeue();
                x = p.X - 1;
                y = p.Y;

                if ( x >= sx && Picture.ColorPoints[x, y].ARGB == white )
                {
                    points.Enqueue( new Imagine.Point( x, y ) );
                    Picture.ColorPoints[x, y] = color;
                }

                x = p.X + 1;

                if ( x < fx && Picture.ColorPoints[x, y].ARGB == white )
                {
                    points.Enqueue( new Imagine.Point( x, y ) );
                    Picture.ColorPoints[x, y] = color;
                }

                x = p.X;
                y = p.Y - 1;

                if ( y >= sy && Picture.ColorPoints[x, y].ARGB == white )
                {
                    points.Enqueue( new Imagine.Point( x, y ) );
                    Picture.ColorPoints[x, y] = color;
                }

                y = p.Y + 1;

                if ( y < fy && Picture.ColorPoints[x, y].ARGB == white )
                {
                    points.Enqueue( new Imagine.Point( x, y ) );
                    Picture.ColorPoints[x, y] = color;
                }
            }
        }

        protected override void OnResize( EventArgs e )
        {
            base.OnResize( e );

            if ( BrushPanel.Controls.Count == 0 )
                return;

            BrushPanel.Hide();
            
            int h = BrushPanel.Height * 6 / BrushPanel.Controls.Count;
            BrushPanel.Controls.Cast<Button>().ForEach( b => b.Height = h );

            BrushPanel.Show();
        }

        private void OnChangePicture( object sender, EventArgs e )
        {
            ChangePicture();
        }

        private void OnUndo( object sender, EventArgs e )
        {
            if ( History.Count == 0 )
                return;

            var undo = History.Pop();
            Fill( undo.Point.X, undo.Point.Y, undo.Color );
        }
        
        private void OnImageClick( object sender, MouseEventArgs e )
        {
#if DEBUG
            if ( e.Button == MouseButtons.Right )
            {
                ChangePicture();
                return;
            }

            if ( e.Button == MouseButtons.Middle )
            {
                ImageFiles.Remove( ImageFile );
                File.Delete( ImageFile );
                ChangePicture();
                return;
            }
#endif
            double scale =
                Math.Max(
                    (double)Picture.Width / ImageBox.Width,
                    (double)Picture.Height / ImageBox.Height );

            double devX = (ImageBox.Width - Picture.Width / scale) / 2;
            double devY = (ImageBox.Height - Picture.Height / scale) / 2;

            int x = (int)(scale * (e.X - devX));
            int y = (int)(scale * (e.Y - devY));
            var white = Picture.ColorPoints[x, y];

            Fill( x, y, PaintColor );
            History.Push( new UndoInfo( white, x, y ) );
        }
        #endregion
    }
}